"""
Decorator for PyPDF2 PdfObject that extracts a couple of properties (type, label, etc).
"""
from typing import Any, List, Optional, Union

from PyPDF2.generic import DictionaryObject, IndirectObject, NumberObject, PdfObject
from rich.markup import escape
from rich.text import Text

from pdfalyzer.helpers.pdf_object_helper import pypdf_class_name
from pdfalyzer.helpers.rich_text_helper import get_type_style, get_type_string_style
from pdfalyzer.helpers.string_helper import root_address
from pdfalyzer.output.layout import get_label_style
from pdfalyzer.output.pdf_node_rich_table import get_node_type_style
from pdfalyzer.util.adobe_strings import *


class PdfObjectProperties:
    """Simple class to extract critical features of a PdfObject."""
    def __init__(
            self,
            pdf_object: PdfObject,
            address: str,
            idnum: int,
            indirect_object: Optional[IndirectObject] = None
        ):
        self.idnum = idnum
        self.obj = pdf_object
        self.indirect_object = indirect_object
        self.sub_type = None
        self.all_references_processed = False
        self.known_to_parent_as: Optional[str] = None

        if isinstance(pdf_object, DictionaryObject):
            self.type = pdf_object.get(TYPE) or address
            self.sub_type = pdf_object.get(SUBTYPE) or pdf_object.get(S)

            if TYPE in pdf_object and self.sub_type is not None:
                self.label = f"{self.type}:{self.sub_type[1:]}"
            else:
                self.label = self.type

            if isinstance(self.type, str):
                self.type = root_address(self.type)
                self.label = root_address(self.label)
        else:
            # If it's not a DictionaryObject all we have as far as naming is the address passed in.
            self.label = address
            self.type = root_address(address) if isinstance(address, str) else None

        # Force a string. TODO this sucks.
        if isinstance(self.label, int):
            self.label = f"{UNLABELED}[{self.label}]"

        # TODO: this is hacky/temporarily incorrect bc we often don't know the parent when node is being constructed
        if isinstance(address, int):
            self.first_address = f"[{address}]"
        else:
            self.first_address = address

    @classmethod
    def from_reference(cls, reference: IndirectObject, address: str) -> 'PdfObjectProperties':
        return cls(reference.get_object(), address, reference.idnum, reference)

    @classmethod
    def resolve_references(cls, reference_key: str, obj: PdfObject) -> Any:
        """Recursively build the same data structure except IndirectObjects are resolved to nodes."""
        if isinstance(obj, NumberObject):
            return obj.as_numeric()
        elif isinstance(obj, IndirectObject):
            return cls.from_reference(obj, reference_key)
        elif isinstance(obj, list):
            return [cls.resolve_references(reference_key, item) for item in obj]
        elif isinstance(obj, dict):
            return {k: cls.resolve_references(k, v) for k, v in obj.items()}
        else:
            return obj

    @classmethod
    def to_table_row(
            cls,
            reference_key: str,
            obj: PdfObject,
            is_single_row_table: bool = False
        ) -> List[Union[Text, str]]:
        """PDF object property at reference_key becomes a formatted 3-tuple for use in Rich tables."""
        return [
            Text(f"{reference_key}", style='grey' if isinstance(reference_key, int) else ''),
            Text(f"{escape(str(cls.resolve_references(reference_key, obj)))}", style=get_type_style(type(obj))),
            # 3rd col (AKA type(value)) is redundant if it's a TextString/Number/etc. node so we make it empty
            '' if is_single_row_table else Text(pypdf_class_name(obj), style=get_type_string_style(type(obj)))
        ]

    def __rich__(self) -> Text:
        return node_label(self.idnum, self.label, self.obj)

    def __str__(self) -> str:
        return self.__rich__().plain


def node_label(idnum: int, label: str, pdf_object: PdfObject) -> Text:
    """Colored text representation of a node."""
    text = Text('<', style='white')
    text.append(f'{idnum}', style='bright_white')
    text.append(':', style='white')
    text.append(label[1:], style=f'{get_label_style(label)} underline bold')
    text.append('(', style='white')
    text.append(pypdf_class_name(pdf_object), style=get_node_type_style(pdf_object))
    text.append(')', style='white')
    text.append('>')
    return text



"""
Class to handle attempting to decode a chunk of bytes into strings with various possible encodings.
Leverages the chardet library to both guide what encodings are attempted as well as to rank decodings
in the results.
"""

from collections import defaultdict
from operator import attrgetter
from typing import List, Optional

from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from pdfalyzer.binary.bytes_match import BytesMatch
from pdfalyzer.binary.decoding_attempt import DecodingAttempt
from pdfalyzer.config import PdfalyzerConfig

from pdfalyzer.detection.encoding_assessment import EncodingAssessment
from pdfalyzer.detection.constants.character_encodings import ENCODING, ENCODINGS_TO_ATTEMPT
from pdfalyzer.detection.encoding_detector import EncodingDetector
from pdfalyzer.helpers.bytes_helper import clean_byte_string
from pdfalyzer.helpers.dict_helper import get_dict_key_by_value
from pdfalyzer.helpers.rich_text_helper import CENTER, DECODING_ERRORS_MSG, NO_DECODING_ERRORS_MSG, console
from pdfalyzer.output.decoding_attempts_table import (DecodingTableRow, assessment_only_row, empty_decoding_attempts_table,
     decoding_table_row)
from pdfalyzer.util.logging import log


# Messages used in the table to show true vs. false (a two element array can be indexed by booleans)
WAS_DECODABLE_YES_NO = [NO_DECODING_ERRORS_MSG, DECODING_ERRORS_MSG]

# Multiply chardet scores by 100 (again) to make sorting the table easy
SCORE_SCALER = 100.0


class BytesDecoder:
    def __init__(self, bytes_match: BytesMatch, label: Optional[str] = None) -> None:
        """Instantiated with _bytes as the whole stream; :bytes_seq tells it how to pull the bytes it will decode"""
        self.bytes_match = bytes_match
        self.bytes = bytes_match.surrounding_bytes
        self.label = label or bytes_match.label

        # Empty table/metrics/etc
        self.table = empty_decoding_attempts_table(bytes_match)
        self.was_match_decodable = _build_encodings_metric_dict()
        self.was_match_force_decoded = _build_encodings_metric_dict()
        self.was_match_undecodable = _build_encodings_metric_dict()
        self.decoded_strings = {}  # dict[encoding: decoded string]
        self.undecoded_rows = []

        # Note we send both the match and surrounding bytes used when detecting the encoding
        self.encoding_detector = EncodingDetector(self.bytes)

    def print_decode_attempts(self) -> None:
        if not PdfalyzerConfig.SUPPRESS_CHARDET_OUTPUT:
            console.print(self.encoding_detector)

        self._print_decode_attempt_subheading()
        console.print(self._generate_decodings_table())

    def _generate_decodings_table(self) -> Table:
        """First rows are the raw / hex views of the bytes"""
        self.decodings = [DecodingAttempt(self.bytes_match, encoding) for encoding in ENCODINGS_TO_ATTEMPT.keys()]

        # Attempt decodings we don't usually attempt if chardet is insistent enough
        forced_decodes = self._undecoded_assessments(self.encoding_detector.force_decode_assessments)
        self.decodings += [DecodingAttempt(self.bytes_match, a.encoding) for a in forced_decodes]

        # If we still haven't decoded chardets top choice, decode it
        if len(self._forced_displays()) > 0 and not self._was_decoded(self._forced_displays()[0].encoding):
            chardet_top_encoding = self._forced_displays()[0].encoding
            log.debug(f"Decoding {chardet_top_encoding} because it's chardet top choice...")
            self.decodings.append(DecodingAttempt(self.bytes_match, chardet_top_encoding))

        rows = [self._row_from_decoding_attempt(decoding) for decoding in self.decodings]
        rows += [assessment_only_row(a, a.confidence * SCORE_SCALER) for a in self._forced_displays()]
        self._track_decode_stats()

        for row in sorted(rows, key=attrgetter('sort_score'), reverse=True):
            self.table.add_row(*row[0:4])

        return self.table

    def _forced_displays(self) -> List[EncodingAssessment]:
        """Returns assessments over the display threshold that are not yet decoded"""
        return self._undecoded_assessments(self.encoding_detector.force_display_assessments)

    def _undecoded_assessments(self, assessments: List[EncodingAssessment]) -> List[EncodingAssessment]:
        """Fiter out the already decoded assessments from a set of assessments"""
        return [a for a in assessments if not self._was_decoded(a.encoding)]

    def _was_decoded(self, encoding: str) -> bool:
        """Check whether a given encoding is in the table already"""
        return any(row.encoding == encoding for row in self.decodings)

    def _print_decode_attempt_subheading(self) -> None:
        """Generate a rich.Panel for decode attempts"""
        headline = Text(f"Found {self.label.lower()} ", style='decode.subheading') + self.bytes_match.__rich__()
        panel = Panel(headline, style='decode.subheading', expand=False)
        console.print(panel, justify=CENTER)

    def _track_decode_stats(self):
        "Track stats about successful vs. forced vs. failed decode attempts"
        for decoding in self.decodings:
            if decoding.failed_to_decode:
                self.was_match_undecodable[decoding.encoding] += 1
                continue

            self.was_match_decodable[decoding.encoding] += 1

            if decoding.was_force_decoded:
                self.was_match_force_decoded[decoding.encoding] += 1

    def _row_from_decoding_attempt(self, decoding: DecodingAttempt) -> 'DecodingTableRow':
        assessment = self.encoding_detector.get_encoding_assessment(decoding.encoding)
        plain_decoded_string = decoding.decoded_string.plain
        sort_score = assessment.confidence * SCORE_SCALER

        # Replace the decoded text with a "same output as X" where X is the encoding that gave the same result
        if plain_decoded_string in self.decoded_strings.values():
            encoding_with_same_output = get_dict_key_by_value(self.decoded_strings, plain_decoded_string)
            display_text = Text('same output as ', style='color(66) dim italic')
            display_text.append(encoding_with_same_output, style='encoding').append('...', style='white')
        else:
            self.decoded_strings[decoding.encoding] = plain_decoded_string
            display_text = decoding.decoded_string

        # Set failures negative, shave off a little for forced decodes
        if decoding.failed_to_decode:
            sort_score = sort_score * -1 - 100
        elif decoding.was_force_decoded:
            sort_score -= 10

        was_forced = WAS_DECODABLE_YES_NO[int(decoding.was_force_decoded)]
        return decoding_table_row(assessment, was_forced, display_text, sort_score)


def _build_encodings_metric_dict():
    """One key for each key in ENCODINGS_TO_ATTEMPT, values are all 0"""
    metrics_dict = defaultdict(lambda: 0)

    for encoding in ENCODINGS_TO_ATTEMPT.keys():
        metrics_dict[encoding] = 0

    return metrics_dict



"""
Methods to build the rich.table used to display decoding attempts of a given bytes array.

Final output should be rich.table of decoding attempts that are sorted like this:

    1. String representation of undecoded bytes is always the first row
    2. Encodings which chardet.detect() ranked as > 0% likelihood are sorted based on that confidence
    3. Then the unchardetectable:
        1. Decodings that were successful, unforced, and new
        2. Decodings that 'successful' but forced
        3. Decodings that were the same as other decodings
        4. Failed decodings
"""

from collections import namedtuple

from rich.table import Table
from rich.text import Text

from pdfalyzer.binary.bytes_match import BytesMatch
from pdfalyzer.detection.encoding_assessment import EncodingAssessment
from pdfalyzer.helpers.bytes_helper import hex_view_of_raw_bytes, rich_text_view_of_raw_bytes
from pdfalyzer.helpers.rich_text_helper import CENTER, FOLD, MIDDLE, RIGHT, na_txt


DECODE_NOT_ATTEMPTED_MSG = Text('(decode not attempted)', style='no_attempt')
HEX = Text('HEX', style='bytes_title')
RAW_BYTES = Text('Raw', style=f"bytes")


def empty_decoding_attempts_table(bytes_match: BytesMatch) -> Table:
    """First rows are the raw / hex views of the bytes"""
    table = Table(show_lines=True, border_style='bytes', header_style='color(101) bold')

    def add_col(title, **kwargs):
        kwargs['justify'] = kwargs.get('justify', CENTER)
        table.add_column(title, overflow=FOLD, vertical=MIDDLE, **kwargs)

    add_col('Encoding', justify=RIGHT, width=12)
    add_col('Detect Odds', width=len('Detect'))
    add_col('Force?', width=len('Force?'))
    add_col('Decoded Output', justify='left')

    na = na_txt(style=HEX.style)
    table.add_row(HEX, na, na, hex_view_of_raw_bytes(bytes_match.surrounding_bytes, bytes_match))
    na = na_txt(style=RAW_BYTES.style)
    table.add_row(RAW_BYTES, na, na, rich_text_view_of_raw_bytes(bytes_match.surrounding_bytes, bytes_match))
    return table


# The confidence and encoding will not be shown in the final display - instead their Text versions are shown
DecodingTableRow = namedtuple(
    'DecodingTableRow',
    [
        'encoding_text',
        'confidence_text',
        'errors_while_decoded',
        'decoded_string',
        'confidence',
        'encoding',
        'sort_score'
    ])


def decoding_table_row(assessment: EncodingAssessment, is_forced: Text, txt: Text, score: float) -> DecodingTableRow:
    """Get a table row for a decoding attempt"""
    return DecodingTableRow(
        assessment.encoding_text,
        assessment.confidence_text,
        is_forced,
        txt,
        assessment.confidence,
        assessment.encoding,
        sort_score=score)


def assessment_only_row(assessment: EncodingAssessment, score) -> DecodingTableRow:
    """Build a row with just chardet assessment data and no actual decoded string"""
    return decoding_table_row(assessment, na_txt(), DECODE_NOT_ATTEMPTED_MSG, score)



"""
Simple class to keep track of regex matches against binary data.  Basically an re.match object with
some (not many) extra bells and whistles, most notably the surrounding_bytes property.

pre_capture_len and post_capture_len refer to the regex sections before and after the capture group,
e.g. a regex like '123(.*)x:' would have pre_capture_len of 3 and post_capture_len of 2.
"""
import re
from typing import Iterator, Optional

from rich.text import Text

from pdfalyzer.config import PdfalyzerConfig
from pdfalyzer.detection.constants.binary_regexes import DANGEROUS_INSTRUCTIONS, CAPTURE_BYTES
from pdfalyzer.helpers.rich_text_helper import BYTES_BRIGHTER, GREY_ADDRESS, prefix_with_plain_text_obj
from pdfalyzer.util.logging import log

# Regex Capture used when extracting quoted chunks of bytes
ALERT_STYLE = 'error'


class BytesMatch:
    def __init__(
            self,
            matched_against: bytes,
            start_idx: int,
            length: int,
            label: str,
            ordinal: int,
            match: Optional[re.Match] = None  # It's rough to get the regex from yara :(
        ) -> None:
        """
        Ordinal means it's the Nth match with this regex (not super important but useful)
        YARA makes it a little rouch to get the actual regex that matched. Can be done with plyara eventually.
        """
        self.matched_against: bytes = matched_against
        self.start_idx: int = start_idx
        self.end_idx: int = start_idx + length
        self.match_length: int = length
        self.length: int = length
        self.label: str = label
        self.ordinal: int = ordinal
        self.match: Optional[re.Match] = match
        # Maybe should be called "matched_bytes"
        self.bytes = matched_against[start_idx:self.end_idx]
        self.match_groups: Optional[tuple] = match.groups() if match else None
        self._find_surrounding_bytes()
        # Adjust the highlighting start point in case this match is very early in the stream
        self.highlight_start_idx = start_idx - self.surrounding_start_idx
        self.highlight_end_idx = self.highlight_start_idx + self.length

        if self.bytes in DANGEROUS_INSTRUCTIONS:
            self.highlight_style = ALERT_STYLE
        else:
            self.highlight_style = BYTES_BRIGHTER

    @classmethod
    def from_regex_match(cls, matched_against: bytes, match: re.Match, ordinal: int) -> 'BytesMatch':
        return cls(matched_against, match.start(), len(match[0]), match.re.pattern, ordinal, match)

    @classmethod
    def from_yara_str(cls, matched_against: bytes, rule_name: str, yara_str: dict, ordinal: int) -> 'BytesMatch':
        """Build a BytesMatch from a yara string match. matched_against is the set of bytes yara was run against"""
        return cls(matched_against, yara_str[0], len(yara_str[2]), rule_name + ' ' + yara_str[1], ordinal)

    @classmethod
    def for_yara_strings_in_match(cls, matched_against: bytes, yara_match: dict) -> Iterator['BytesMatch']:
        """
        Iterator over all strings returned as part of a yara match dict, which looks like this:
        {
            'tags': ['foo', 'bar'],
            'matches': True,
            'namespace': 'default',
            'rule': 'my_rule',
            'meta': {},
            'strings': [(81L, '$a', 'abc'), (141L, '$b', 'def')]
        }
        """
        for i, yara_str in enumerate(yara_match['strings']):
            yield(cls.from_yara_str(matched_against, yara_match['rule'], yara_str, i + 1))

    def style_at_position(self, idx) -> str:
        """Get the style for the byte at position idx within the matched bytes"""
        if idx < self.highlight_start_idx or idx >= self.highlight_end_idx:
            return GREY_ADDRESS
        else:
            return self.highlight_style

    def _find_surrounding_bytes(self, num_before: Optional[int] = None, num_after: Optional[int] = None) -> None:
        """Find the surrounding bytes, making sure not to step off the beginning or end"""
        num_after = num_after or num_before or PdfalyzerConfig.NUM_SURROUNDING_BYTES
        num_before = num_before or PdfalyzerConfig.NUM_SURROUNDING_BYTES
        self.surrounding_start_idx: int = max(self.start_idx - num_before, 0)
        self.surrounding_end_idx: int = min(self.end_idx + num_after, len(self.matched_against))
        self.surrounding_bytes: bytes = self.matched_against[self.surrounding_start_idx:self.surrounding_end_idx]

    def __rich__(self) -> Text:
        headline = prefix_with_plain_text_obj(str(self.match_length), style='number', root_style='decode.subheading')
        headline.append(f" bytes matching ")
        headline.append(f"{self.label} ", style=ALERT_STYLE if self.highlight_style == ALERT_STYLE else 'regex')
        headline.append('at ')
        headline.append(f"(start idx: ", style='off_white')
        headline.append(str(self.start_idx), style='number')
        headline.append(', end idx: ', style='off_white')
        headline.append(str(self.end_idx), style='number')
        headline.append(')', style='off_white')
        return headline

    def __str__(self):
        return self.__rich__().plain



"""
Walks the PDF objects and builds the PDF logical structure tree by
wrapping each internal PDF object in a PdfTreeNode. Tree is managed by
the anytree library. Information about the tree as a whole is stored
in this class.
Once the PDF is parsed this class manages access to
information about or from the underlying PDF tree.
"""
from os.path import basename
from typing import Dict, Iterator, List, Optional

from anytree import LevelOrderIter, SymlinkNode
from anytree.search import findall, findall_by_attr
from PyPDF2 import PdfReader
from PyPDF2.errors import PdfReadError
from PyPDF2.generic import IndirectObject, NameObject, NumberObject
from rich.markup import escape
from yaralyzer.helpers.bytes_helper import get_bytes_info
from yaralyzer.helpers.file_helper import load_binary_data
from yaralyzer.output.rich_console import console
from yaralyzer.util.logging import log, set_log_level

from pdfalyzer.decorators.document_model_printer import print_with_header
from pdfalyzer.decorators.pdf_tree_node import PdfTreeNode, find_node_with_most_descendants
from pdfalyzer.font_info import FontInfo
from pdfalyzer.helpers.string_helper import all_strings_are_same_ignoring_numbers, has_a_common_substring
from pdfalyzer.pdf_object_relationship import PdfObjectRelationship
from pdfalyzer.util.adobe_strings import *
from pdfalyzer.util.exceptions import PdfWalkError

TRAILER_FALLBACK_ID = 10000000


class Pdfalyzer:
    def __init__(self, pdf_path: str):
        self.pdf_path = pdf_path
        self.pdf_basename = basename(pdf_path)
        self.pdf_bytes = load_binary_data(pdf_path)
        self.pdf_bytes_info = get_bytes_info(self.pdf_bytes)
        pdf_file = open(pdf_path, 'rb')  # Filehandle must be left open for PyPDF2 to perform seeks
        self.pdf_reader = PdfReader(pdf_file)

        # Initialize tracking variables
        self.indeterminate_ids = set()  # See INDETERMINATE_REF_KEYS comment
        self.traversed_nodes: Dict[int, PdfTreeNode] = {}  # Nodes we've seen already
        self.font_infos: List[FontInfo] = []  # Font summary objects
        self.max_generation = 0  # PDF revisions are "generations"; this is the max generation encountered

        # Bootstrap the root of the tree with the trailer. PDFs are always read trailer first.
        # Extract trailer. Technically the trailer has no PDF obj. ID but we set it to the /Size of the PDF
        trailer = self.pdf_reader.trailer
        self.pdf_size = trailer.get(SIZE)
        trailer_id = self.pdf_size if self.pdf_size is not None else TRAILER_FALLBACK_ID
        self.pdf_tree = PdfTreeNode(trailer, TRAILER, trailer_id)
        self.traversed_nodes[self.pdf_tree.idnum] = self.pdf_tree
        self.walk_node(self.pdf_tree)  # Build tree by recursively following relationships between nodes

        # After scanning all objects we place nodes whose position was uncertain, extract fonts, and verify
        self._resolve_indeterminate_nodes()
        self._extract_font_infos()
        self._verify_all_traversed_nodes_are_in_tree()
        self._verify_untraversed_nodes_are_untraversable()

        # Create SymlinkNodes for relationships between PDF objects that are not parent/child relationships.
        # (Do this last because it has the side effect of making a lot more nodes)
        for node in self.node_iterator():
            if not isinstance(node, SymlinkNode):
                node.symlink_non_tree_relationships()

        log.info(f"Walk complete.")

    def walk_node(self, node: PdfTreeNode) -> None:
        """Recursively walk the PDF's tree structure starting at a given node"""
        log.info(f'walk_node() called with {node}. Object dump:\n{print_with_header(node.obj, node.label)}')
        nodes_to_walk_next = [self._process_relationship(r) for r in node.references_to_other_nodes()]
        node.all_references_processed = True

        for next_node in [n for n in nodes_to_walk_next if not (n is None or n.all_references_processed) ]:
            if not next_node.all_references_processed:
                self.walk_node(next_node)

    def find_node_by_idnum(self, idnum) -> Optional[PdfTreeNode]:
        """Find node with idnum in the tree. Return None if that node is not reachable from the root."""
        nodes = [
            node for node in findall_by_attr(self.pdf_tree, name='idnum', value=idnum)
            if not isinstance(node, SymlinkNode)
        ]

        if len(nodes) == 0:
            return None
        elif len(nodes) == 1:
            return nodes[0]
        else:
            raise PdfWalkError(f"Too many nodes had id {idnum}: {nodes}")

    def stream_nodes(self) -> List[PdfTreeNode]:
        """List of actual nodes (not SymlinkNodes) containing streams sorted by PDF object ID"""
        stream_filter = lambda node: node.contains_stream() and not isinstance(node, SymlinkNode)
        return sorted(findall(self.pdf_tree, stream_filter), key=lambda r: r.idnum)

    def is_in_tree(self, search_for_node: PdfTreeNode) -> bool:
        """Returns true if search_for_node is in the tree already."""
        for node in self.node_iterator():
            if node == search_for_node:
                return True

        return False

    def node_iterator(self) -> Iterator[PdfTreeNode]:
        """Iterate over nodes, grouping them by distance from the root."""
        return LevelOrderIter(self.pdf_tree)

    def _process_relationship(self, relationship: PdfObjectRelationship) -> Optional[PdfTreeNode]:
        """
        Place the relationship 'node' in the tree. Returns a list of nodes to walk next.
        'address' is the key used in node.obj to refer to 'relationship' object
           plus any modifiers like [2] or [/Something]
        """
        log.info(f'Assessing relationship {relationship}...')
        was_seen_before = (relationship.to_obj.idnum in self.traversed_nodes) # Must come before _build_or_find()
        from_node = relationship.from_node
        to_node = self._build_or_find_node(relationship.to_obj, relationship.address)
        self.max_generation = max([self.max_generation, relationship.to_obj.generation or 0])

        # If one is already a parent/child of the other there's nothing to do
        if to_node == from_node.parent or to_node in from_node.children:
            log.debug(f"  {from_node} and {to_node} are already related")
            return None

        # Many branches return None or don't return.
        # If there's an explicit /Parent or /Kids relationship then we know the correct relationship
        if relationship.is_parent or relationship.is_child:
            log.debug(f"  Explicit parent/child link: {relationship}")

            if relationship.is_parent:
                from_node.set_parent(to_node)
            elif to_node.parent is not None:
                if relationship.from_node.type.startswith(STRUCT_ELEM):# reference_key != relationship.address:
                    log.info(f"{relationship} to {to_node} has different key v. address. Not placed; parent is {to_node.parent}")
                else:
                    log.warning(f"{relationship} cound not become parent of {to_node} bc parent is {to_node.parent}")
            else:
                from_node.add_child(to_node)

            # Remove this to_node from inteterminacy now that it's got a child or parent
            if relationship.to_obj.idnum in self.indeterminate_ids:
                log.info(f"  Found {relationship} => {to_node} was marked indeterminate but now placed")
                self.indeterminate_ids.remove(relationship.to_obj.idnum)
        elif relationship.is_indeterminate or relationship.is_link or was_seen_before:
            to_node.add_non_tree_relationship(relationship)

            if was_seen_before:
                if relationship.to_obj.idnum not in self.indeterminate_ids and to_node.parent is None:
                    raise PdfWalkError(f"{relationship} - ref has no parent and is not indeterminate")
                else:
                    log.debug(f"  Already saw {relationship}; not scanning next")
                    return None

            # Indeterminate relationships need to wait until everything has been scanned to be placed
            if relationship.is_indeterminate or (relationship.is_link and not self.is_in_tree(to_node)):
                log.info(f'  Indeterminate ref {relationship}')
                self.indeterminate_ids.add(to_node.idnum)
            elif relationship.is_link:
                log.debug(f"  Link ref {relationship}")  # Link nodes like /Dest are usually just links between nodes
        else:
            # If no other conditions are met, add the relationship as a child
            from_node.add_child(to_node)

        return to_node

    def _resolve_indeterminate_nodes(self) -> None:
        """
        Some nodes cannot be placed until we have walked the rest of the tree. For instance
        if we encounter a /Page that relationships /Resources we need to know if there's a
        /Pages parent of the /Page before committing to a tree structure.
        """
        #set_log_level('INFO')
        indeterminate_nodes = [self.traversed_nodes[idnum] for idnum in self.indeterminate_ids]
        indeterminate_nodes_string = "\n   ".join([f"{node}" for node in indeterminate_nodes])
        log.info(f"Resolving {len(indeterminate_nodes)} indeterminate nodes: {indeterminate_nodes_string}")

        for node in indeterminate_nodes:
            # We're basically checking for situations where we can just pick the lowest ID object that
            # has a relationship to the indeterminate node. In rare cases we will override 'possible_parent_relationships'.
            if node.parent is not None:
                log.info(f"{node} marked indeterminate but has parent: {node.parent}")
                continue
            elif self._attempt_tree_placement(node):
                log.info("SUCCESSFULLY PLACED indeterminate node custom style")
                continue

            log.debug(f"Attempting to resolve indeterminate node: {node}")
            possible_parent_relationships = node.non_tree_relationships.copy()
            unique_refferer_labels = node.unique_labels_of_referring_nodes()
            unique_addresses = node.unique_addresses()

            # Check addresses and referring node labels to see if they are all the same
            reference_keys_or_nodes_are_same = any([
                all_strings_are_same_ignoring_numbers(_list) or has_a_common_substring(_list)
                for _list in [unique_addresses, unique_refferer_labels]
            ])

            # Clauses that don't 'raise' or 'continue' result in related node with lowest ID being made parent
            if reference_keys_or_nodes_are_same:
                log.info(f"Fuzzy match addresses: {unique_addresses} / labels: {unique_refferer_labels}")
            elif node.type == COLOR_SPACE:
                log.info("Color space node found; placing at lowest ID")
            else:
                log.warning(f"Indeterminate {node} parent based on descendant count, not PDF logic. Options were:")
                node.log_non_tree_relationships()

            possible_parents = [r.from_node for r in possible_parent_relationships]
            node.set_parent(find_node_with_most_descendants(possible_parents))

    def _attempt_tree_placement(self, node: PdfTreeNode) -> bool:
        """Attempt to find place for node in self.pdf_tree."""
        # TODO we should have removed the node from indeterminate_node_ids before this point
        common_ancestor_among_possible_parents = node.common_ancestor_among_non_tree_relationships()

        # As opposed to INDETERMINITE
        determinate_relationships = [
            r for r in node.non_tree_relationships
            if r.from_node.type not in NON_TREE_KEYS
        ]

        log.info(f"Found {len(determinate_relationships)} determinate relationships...")

        for r in determinate_relationships:
            log.debug(f"  Determinate relationship: {r}")

        if common_ancestor_among_possible_parents is not None:
            log.info(f"  Found common ancestor: {common_ancestor_among_possible_parents}")
            node.set_parent(common_ancestor_among_possible_parents)
        elif len(determinate_relationships) == 1:
            log.info(f"  Single determinate_relationship {determinate_relationships[0]}; making it the parent")
            node.set_parent(determinate_relationships[0].from_node)
        elif node.label.startswith(RESOURCES) and node.unique_labels_of_referring_nodes() == set([PAGE, PAGES]):
            # An edge case seen in the wild involving a PDF that doesn't conform to the PDF spec
            log.warning(f"  Failed to place {node}; seems to be a loose {PAGE}. Linking to first {PAGES}")
            pages_nodes = [n for n in node.nodes_with_non_tree_references_to_self() if node.type == PAGES]
            node.set_parent(sorted(pages_nodes, key=lambda n: n.idnum)[0])
        else:
            # Then we didn't manage to place it.
            return False

        return True

    def _extract_font_infos(self) -> None:
        """Extract information about fonts in the tree and place it in self.font_infos"""
        for node in self.node_iterator():
            if isinstance(node.obj, dict) and RESOURCES in node.obj:
                log.debug(f"Extracting fonts from node with '{RESOURCES}' key: {node}...")
                known_font_ids = [fi.idnum for fi in self.font_infos]

                self.font_infos += [
                    fi for fi in FontInfo.extract_font_infos(node.obj)
                    if fi.idnum not in known_font_ids
                ]

    def _build_or_find_node(self, relationship: IndirectObject, relationship_key: str) -> PdfTreeNode:
        """If node exists in self.traversed_nodes return it, otherwise build a node and store it."""
        if relationship.idnum in self.traversed_nodes:
            return self.traversed_nodes[relationship.idnum]

        log.debug(f"Building node for {relationship}")
        new_node = PdfTreeNode.from_reference(relationship, relationship_key)
        self.traversed_nodes[relationship.idnum] = new_node
        return new_node

    def _print_traversed_nodes(self) -> None:
        """Debug method that displays which nodes have already been walked"""
        for i in sorted(self.traversed_nodes.keys()):
            console.print(f'{i}: {self.traversed_nodes[i]}')

    def _verify_all_traversed_nodes_are_in_tree(self) -> None:
        """Make sure every node we can see is reachable from the root of the tree"""
        missing_nodes = [
            node for idnum, node in self.traversed_nodes.items()
            if self.find_node_by_idnum(idnum) is None
        ]

        if len(missing_nodes) > 0:
            msg = f"Nodes were traversed but never placed: {escape(str(missing_nodes))}\n" + \
                   "For link nodes like /First, /Next, /Prev, and /Last this might be no big deal - depends " + \
                   "on the PDF. But for other node typtes this could indicate missing data in the tree."
            console.print(msg)
            log.warning(msg)

    def _verify_untraversed_nodes_are_untraversable(self) -> None:
        """Make sure any PDF object IDs we can't find in tree are /ObjStm or /Xref nodes"""
        if self.pdf_size is None:
            log.warning(f"{SIZE} not found in PDF trailer; cannot verify all nodes are in tree")
            return
        if self.max_generation > 0:
            log.warning(f"_verify_untraversed_nodes_are_untraversable() only checking generation {self.max_generation}")

        # We expect to see all ordinals up to the number of nodes /Trailer claims exist as obj. IDs.
        missing_node_ids = [i for i in range(1, self.pdf_size) if self.find_node_by_idnum(i) is None]

        for idnum in missing_node_ids:
            ref = IndirectObject(idnum, self.max_generation, self.pdf_reader)

            try:
                obj = ref.get_object()
            except PdfReadError as e:
                if 'Invalid Elementary Object' in str(e):
                    log.warning(f"Couldn't verify elementary obj with id {idnum} is properly in tree")
                    continue
                log.error(str(e))
                console.print_exception()
                obj = None
                raise e

            if obj is None:
                log.error(f"Cannot find ref {ref} in PDF!")
                continue
            elif isinstance(obj, (NumberObject, NameObject)):
                log.info(f"Obj {idnum} is a {type(obj)} w/value {obj}; if relationshipd by /Length etc. this is a nonissue but maybe worth doublechecking")
                continue
            elif not isinstance(obj, dict):
                log.error(f"Obj {idnum} ({obj}) of type {type(obj)} isn't dict, cannot determine if it should be in tree")
                continue
            elif TYPE not in obj:
                msg = f"Obj {idnum} has no {TYPE} and is not in tree. Either a loose node w/no data or an error in pdfalyzer."
                msg += f"\nHere's the contents for you to assess:\n{obj}"
                log.warning(msg)
                continue

            obj_type = obj[TYPE]

            if obj_type == OBJECT_STREAM:
                log.debug(f"Object with id {idnum} not found in tree because it's an {OBJECT_STREAM}")
            elif obj[TYPE] == XREF:
                placeable = XREF_STREAM in self.pdf_reader.trailer

                for k, v in self.pdf_reader.trailer.items():
                    xref_val_for_key = obj.get(k)

                    if k in [XREF_STREAM, PREV]:
                        continue
                    elif k == SIZE:
                        if xref_val_for_key is None or v != (xref_val_for_key + 1):
                            log.info(f"{XREF} has {SIZE} of {xref_val_for_key}, trailer has {SIZE} of {v}")
                            placeable = False

                        continue
                    elif k not in obj or v != obj.get(k):
                        log.info(f"Trailer has {k} -> {v} but {XREF} obj has {obj.get(k)} at that key")
                        placeable = False

                if placeable:
                    self.pdf_tree.add_child(self._build_or_find_node(ref, XREF_STREAM))
            else:
                log.warning(f"{XREF} Obj {idnum} not found in tree!")



"""
Methods to help with the design of the output
"""
import re

from rich.panel import Panel
from yaralyzer.output.rich_console import console, console_width

from pdfalyzer.util import adobe_strings

HEADER_PADDING = (1, 1)
DEFAULT_LABEL_STYLE = 'yellow'
FONT_OBJ_BLUE = 'deep_sky_blue4 bold'
PDF_NON_TREE_REF = 'color(243)'

LABEL_STYLES = [
    [re.compile('JavaScript|JS|OpenAction', re.I | re.M), 'blink bold red'],
    [re.compile(f'^{adobe_strings.FONT_DESCRIPTOR}'),     'cornflower_blue'],
    [re.compile(f'^{adobe_strings.FONT_FILE}'),           'steel_blue1'],
    [re.compile(f'^{adobe_strings.FONT}'),                FONT_OBJ_BLUE],
    [re.compile(f'^{adobe_strings.TO_UNICODE}'),          'grey30'],
    [re.compile(f'^{adobe_strings.WIDTHS}'),              'color(67)'],
    [re.compile(f'^{adobe_strings.W}'),                   'color(67)'],
    [re.compile(f'^{adobe_strings.RESOURCES}'),           'magenta'],
    [re.compile('/(Trailer|Root|Info|Outlines)'),         'bright_green'],
    [re.compile('/Catalog'),                              'color(47)'],
    [re.compile('/(Metadata|ViewerPreferences)'),         'color(35)'],
    [re.compile('^/Contents'),                            'medium_purple1'],
    [re.compile('^/Action'),                              'dark_red'],
    [re.compile('^/Annots'),                              'deep_sky_blue4'],
    [re.compile('^/Annot'),                               'color(24)'],
    [re.compile('^/Pages'),                               'dark_orange3'],
    [re.compile('^/Page'),                                'light_salmon3'],
    [re.compile('^/ColorSpace'),                          'medium_orchid1'],
    [re.compile('^/(URI|Names)'),                         'white'],
    [re.compile(f'^{adobe_strings.XOBJECT}'),             'grey37'],
    [re.compile(f'^{adobe_strings.UNLABELED}'),           'grey35 reverse'],
    [re.compile(f'^{adobe_strings.XREF}'),                'color(148)'],
]

LABEL_STYLES += [
    [re.compile(f'^{key}'), PDF_NON_TREE_REF]
    for key in adobe_strings.NON_TREE_REFERENCES
]


def subheading_width() -> int:
    return int(console_width() * 0.75)


def half_width() -> int:
    return int(console_width() * 0.5)


def print_section_header(headline: str, style: str = '') -> None:
    console.line(2)
    _print_header_panel(headline, f"{style} reverse", True, console_width(), HEADER_PADDING)
    console.line()


def print_section_subheader(headline: str, style: str = '') -> None:
    console.line()
    _print_header_panel(headline, style, True, subheading_width(), HEADER_PADDING)


def print_section_sub_subheader(headline: str, style: str = ''):
    console.line()
    _print_header_panel(headline, style, True, half_width())


def print_headline_panel(headline, style: str = ''):
    _print_header_panel(headline, style, False, console_width())


def _print_header_panel(headline: str, style: str, expand: bool, width: int, padding: tuple = (0,)) -> None:
    console.print(Panel(headline, style=style, expand=expand, width=width or subheading_width(), padding=padding))


def get_label_style(label: str) -> str:
    """Lookup a style based on the label string"""
    return next((ls[1] for ls in LABEL_STYLES if ls[0].search(label)), DEFAULT_LABEL_STYLE)



from datetime import datetime


def timestamp_for_filename() -> str:
    """Returns a string showing current time in a file name friendly format"""
    return datetime.now().strftime("%Y-%m-%dT%H.%M.%S")


def load_word_list(file_path):
    """For very simple files (1 col CSVs, if you wll)"""
    with open(file_path, 'r') as f:
        return [line.rstrip().lstrip() for line in f.readlines()]


def load_binary_data(file_path):
    with open(file_path, 'rb') as f:
        return f.read()



"""
Various text formatting/styling/manipulating methods.
"""
import re
from pprint import PrettyPrinter
from typing import List, Pattern, Union

from PyPDF2.generic import PdfObject
from yaralyzer.output.rich_console import console_width
from yaralyzer.util.logging import log

INDENT_DEPTH = 4
PRETTY_PRINT_WIDTH = 60
DIGIT_REGEX = re.compile("\\d+")

# Pretty Printer
pp = PrettyPrinter(
    indent=INDENT_DEPTH,
    width=PRETTY_PRINT_WIDTH,
    sort_dicts=True)


def generate_hyphen_line(width=None, title=None):
    """e.g. '-----------------BEGIN-----------------'"""
    width = width or console_width()

    if title is None:
        return '-' * width

    side_hyphens = int((width - len(title)) / 2) * '-'
    line = side_hyphens + title + side_hyphens
    return line if len(line) == width else line + '-'


def count_pattern_matches_in_text(pattern: str, text: str) -> int:
    return count_regex_matches_in_text(re.compile(pattern), text)


def count_regex_matches_in_text(regex: Pattern, text: str) -> int:
    """For use when you precompile the regex"""
    return sum(1 for _ in regex.finditer(text))


def root_address(_string: str) -> str:
    """Strip the bracketed part off an address, e.g. '/Root[1]' => '/Root'."""
    return _string.split('[')[0]


def is_prefixed_by_any(_string: str, prefixes: List[str]) -> bool:
    """Returns True if _string starts with anything in 'prefixes'."""
    return any([_string.startswith(prefix) for prefix in prefixes])


def bracketed(index: Union[int, str]) -> str:
    """Surround index with [ and ]."""
    return f"[{index}]"


def replace_digits(string_with_digits: str) -> str:
    """Turn all digits to X chars in a string."""
    return DIGIT_REGEX.sub('x', string_with_digits)


def all_strings_are_same_ignoring_numbers(strings: List[str]) -> bool:
    """Returns true if string addresses are same except for digits."""
    return len(set([replace_digits(s) for s in strings])) == 1


def is_substring_of_longer_strings_in_list(_string: str, strings: List[str]) -> bool:
    longer_strings = [s for s in strings if len(s) > len(_string)]
    return all([_string in longer_string for longer_string in longer_strings])


def has_a_common_substring(strings: List[str]) -> bool:
    return all([
        is_substring_of_longer_strings_in_list(s, strings)
        for s in strings
    ])



"""
Class to smooth some of the rough edges around the dicts returned by chardet.detect_all()
"""
from typing import Any

from rich.text import Text

from pdfalyzer.detection.constants.character_encodings import ENCODING
from pdfalyzer.helpers.rich_text_helper import (DIM_COUNTRY_THRESHOLD, console, meter_style,
     prefix_with_plain_text_obj)

CONFIDENCE = 'confidence'
LANGUAGE = 'language'


class EncodingAssessment:
    def __init__(self, assessment: dict) -> None:
        self.assessment = assessment
        self.encoding = assessment[ENCODING].lower()
        self.encoding_text = Text(self.encoding, 'encoding.header')
        self.language = self._get_dict_empty_value_as_None(LANGUAGE)
        self.language_text = None if self.language is None else Text(self.language, 'encoding.language')

        # Shift confidence from 0-1.0 scale to 0-100.0 scale
        confidence = self._get_dict_empty_value_as_None(CONFIDENCE) or 0.0
        assert isinstance(confidence, float)
        self.confidence = 100.0 * confidence
        self.confidence_text = prefix_with_plain_text_obj(f"{round(self.confidence, 1)}%", style=meter_style(self.confidence))

        # Pair the language name with the encoding name into one Text obj
        if self.language is not None:
            dim = 'dim' if confidence < DIM_COUNTRY_THRESHOLD else ''
            self.encoding_text.append(f" ({self.language.title()})", style=f"color(23) {dim}")

    @classmethod
    def dummy_encoding_assessment(cls, encoding) -> 'EncodingAssessment':
        """Generate an empty EncodingAssessment to use as a dummy when chardet gives us nothing"""
        assessment = cls({ENCODING: encoding, 'confidence': 0.0})
        assessment.confidence_text = Text('none', 'no_attempt')
        return assessment

    def __rich__(self) -> Text:
        return Text('<Chardet(', 'white') + self.encoding_text + Text(':') + self.confidence_text + Text('>')

    def __str__(self) -> str:
        return self.__rich__().plain

    def _get_dict_empty_value_as_None(self, key: str) -> Any:
        """Return None if the value at :key is an empty string, empty list, etc."""
        value = self.assessment.get(key)

        if isinstance(value, (dict, list, str)) and len(value) == 0:
            return None
        else:
            return value



"""
Class to  measure what we enounter as we iterate over every single match of a relatively simple byte level regex
(e.g. "bytes between quotes") against a relatively large pool of close to random encrypted binary data

Things like how much many of our matched bytes were we able to decode easily vs. by force vs. not at all,
were some encodings have a higher pct of success than others (indicating part of our mystery data might be encoded
that way?
"""
from collections import defaultdict

from pdfalyzer.detection.constants.character_encodings import ENCODINGS_TO_ATTEMPT


class RegexMatchMetrics:
    def __init__(self) -> None:
        self.match_count = 0
        self.bytes_matched = 0
        self.matches_decoded = 0
        self.easy_decode_count = 0
        self.forced_decode_count = 0
        self.skipped_matches_lengths = defaultdict(lambda: 0)
        self.was_match_decodable = defaultdict(lambda: 0)
        self.was_match_force_decoded = defaultdict(lambda: 0)
        self.was_match_undecodable = defaultdict(lambda: 0)
        self.bytes_match_objs = []

    def num_matches_skipped_for_being_empty(self) -> int:
        return self.skipped_matches_lengths[0]

    def num_matches_skipped_for_being_too_big(self) -> int:
        return sum({k: v for k, v in self.skipped_matches_lengths.items() if k > 0}.values())

    def __str__(self):
        return f"<matches: {self.match_count}, " + \
               f"bytes: {self.bytes_matched}, " + \
               f"decoded: {self.matches_decoded} " + \
               f"too_big: {self.num_matches_skipped_for_being_too_big()}, " + \
               f"empty: {self.num_matches_skipped_for_being_empty()}>"



from sys import byteorder
from typing import Optional

from rich.markup import escape
from rich.panel import Panel
from rich.text import Text

from pdfalyzer.binary.bytes_match import ALERT_STYLE, BytesMatch
from pdfalyzer.detection.constants.character_encodings import (ENCODINGS_TO_ATTEMPT, SINGLE_BYTE_ENCODINGS,
     UTF_8, UTF_16, UTF_32)
from pdfalyzer.detection.javascript_hunter import JavascriptHunter
from pdfalyzer.helpers.bytes_helper import clean_byte_string
from pdfalyzer.helpers.rich_text_helper import (BYTES_BRIGHTER, BYTES_BRIGHTEST, BYTES_NO_DIM, GREY_ADDRESS,
     prefix_with_plain_text_obj, unprintable_byte_to_text)
from pdfalyzer.util.logging import log


class DecodingAttempt:
    def __init__(self, bytes_match: BytesMatch, encoding: str) -> None:
        self.bytes_match = bytes_match
        self.encoding = encoding
        self.bytes = bytes_match.surrounding_bytes
        self.was_force_decoded = False
        self.failed_to_decode = False
        self.decoded_string = self._decode_bytes()
        self.javascript_keyword_substrings = JavascriptHunter.js_keyword_matches(self.decoded_string.plain)
        self.javascript_keyword_count = len(self.javascript_keyword_substrings)

    def _decode_bytes(self) -> Text:
        """
        Sets self.decoded_string. Tries builtin decode, hands off to other methods for harsher treatement
        (Byte shifting for UTF-16/32 and custom decode for the rest) if that fails
        """
        try:
            decoded_string = self._to_rich_text(escape(self.bytes.decode(self.encoding)))
            log.info(f"{self.encoding} auto-decoded {self.bytes_match}")
            return decoded_string
        except UnicodeDecodeError:
            log.info(f"{self.encoding} failed on 1st pass decoding {self.bytes_match} capture; custom decoding...")
        except LookupError as e:
            log.warn(f"Unknown encoding: {self.encoding}")
            return self._failed_to_decode(e)

        self.was_force_decoded = True

        if self.encoding in [UTF_16, UTF_32]:
            return self._decode_utf_multibyte_with_byte_offset()
        else:
            return self._custom_decode()

    def _custom_decode(self) -> Text:
        """Returns a Text obj representing an attempt to force a UTF-8 encoding upon an array of bytes"""
        log.info(f"Custom decoding {self.bytes_match} with {self.encoding}...")
        unprintable_char_map = ENCODINGS_TO_ATTEMPT.get(self.encoding)
        output = Text('', style='bytes.decoded')

        # We use this to skip over bytes consumed by multi-byte UTF-n chars
        skip_next = 0

        for i, b in enumerate(self.bytes):
            if skip_next > 0:
                skip_next -= 1
                continue

            _byte = b.to_bytes(1, byteorder)

            # Color the before and after bytes grey
            if i < self.bytes_match.highlight_start_idx or i > self.bytes_match.highlight_end_idx:
                style = GREY_ADDRESS
            else:
                style = self.bytes_match.highlight_style

            if style not in [GREY_ADDRESS, ALERT_STYLE]:
                if b <= 126:
                    style = BYTES_NO_DIM
                elif b <= 192:
                    style = BYTES_BRIGHTER
                else:
                    style = BYTES_BRIGHTEST

            try:
                if unprintable_char_map is not None and b in unprintable_char_map:
                    output.append(unprintable_byte_to_text(unprintable_char_map[b], style=style))
                elif b < 127:
                    output.append(_byte.decode(self.encoding), style=style)
                elif self.encoding != UTF_8:
                    output.append(_byte.decode(self.encoding), style=style)
                # At this point we know it's UTF-8, so it must be a continuation byte
                elif b <= 192:
                    # In UTF-8 bytes from 128 to 192 is a continuation byte
                    output.append(unprintable_byte_to_text(f"CHAR{b}", style=style))
                else:
                    if b <= 223:
                        char_width = 2
                    elif b <= 239:
                        char_width = 3
                    else:
                        char_width = 4

                    wide_char = self.bytes[i:i + char_width].decode(self.encoding)
                    output.append(wide_char, style=style)
                    skip_next = char_width - 1  # Won't be set if there's a decoding exception
                    log.info(f"Skipping next {skip_next} bytes because UTF-8 multibyte char '{wide_char}' used them")
            except UnicodeDecodeError:
                output.append(clean_byte_string(_byte), style=style)

        return output

    def _decode_utf_multibyte_with_byte_offset(self) -> Text:
        """ UTF-16/32 are fixed width (and wide)"""
        char_width = 2 if self.encoding == UTF_16 else 4
        log.debug(f"Decoding {self.encoding}, char_width is {char_width}...")
        last_exception = None
        decoded_str = None
        bytes_offset = 1

        while bytes_offset < char_width:
            try:
                decoded_str = self.bytes[bytes_offset:].decode(self.encoding)
            except UnicodeDecodeError as e:
                log.info(f"Exception decoding w/offset {bytes_offset} in {self.encoding}: {e}")
                last_exception = e

            if decoded_str is not None:
                break

            bytes_offset += 1

        if decoded_str is None:
            return self._failed_to_decode(last_exception)

        return self._to_rich_text(decoded_str, bytes_offset)

    def _failed_to_decode(self, exception: Optional[Exception]) -> Text:
        self.failed_to_decode = True
        return prefix_with_plain_text_obj(f"(decode failed: {exception})", style='red dim italic')

    def _to_rich_text(self, _string: str, bytes_offset: int=0) -> Text:
        """Convert a decoded string to highlighted Text representation"""
        # Adjust where we start the highlighting given the multibyte nature of the encodings
        highlight_byte_start_idx = self.bytes_match.highlight_start_idx + bytes_offset
        txt = Text('', style=self.bytes_match.style_at_position(0))
        current_byte_idx = 0
        log.debug(f"Stepping through {self.encoding} encoded string...")

        # Prevent unprintable chars other than newline. Some of them disfigure the terminal output permanently
        if self.encoding in SINGLE_BYTE_ENCODINGS:
            is_single_byte_encoding = True
            unprintable_chars = ENCODINGS_TO_ATTEMPT[self.encoding]
        else:
            is_single_byte_encoding = False
            unprintable_chars = {}

        for i, c in enumerate(_string):
            char_bytes = bytes(c, self.encoding)
            char_width = len(char_bytes)
            style = self.bytes_match.style_at_position(current_byte_idx + bytes_offset)

            # 10 is newline in single byte encodings
            if c.isprintable() or (ord(c) == 10 and is_single_byte_encoding):
                txt.append(c, style)
            elif ord(c) == 9 and is_single_byte_encoding:
                txt.append(unprintable_byte_to_text('\\t', style=style))
            elif ord(c) in unprintable_chars:
                txt.append(unprintable_byte_to_text(unprintable_chars[ord(c)], style=style))
            else:
                txt.append(unprintable_byte_to_text(f"CHAR{ord(c)}", style=style))

            current_byte_idx += char_width

        return txt



"""
Unify font information spread across a bunch of PdfObjects (Font, FontDescriptor,
and FontFile) into a single class.
"""
from typing import Union

from PyPDF2._cmap import build_char_map, prepare_cm
from PyPDF2.generic import IndirectObject, PdfObject
from rich.columns import Columns
from rich.padding import Padding
from rich.table import Table
from rich.text import Text
from yaralyzer.helpers.bytes_helper import print_bytes
from yaralyzer.output.rich_console import console
from yaralyzer.output.rich_layout_elements import bytes_hashes_table
from yaralyzer.util.logging import log

from pdfalyzer.binary.binary_scanner import BinaryScanner
from pdfalyzer.output.layout import get_label_style
from pdfalyzer.helpers.rich_text_helper import get_type_style
from pdfalyzer.helpers.string_helper import pp
from pdfalyzer.output.layout import print_section_subheader, print_headline_panel, subheading_width
from pdfalyzer.util.adobe_strings import (FONT, FONT_DESCRIPTOR, FONT_FILE, FONT_LENGTHS, RESOURCES, SUBTYPE,
     TO_UNICODE, TYPE, W, WIDTHS)

CHARMAP_TITLE = 'Character Mapping (As Extracted By PyPDF2)'
CHARMAP_TITLE_PADDING = (1, 0, 0, 2)
CHARMAP_PADDING = (0, 2, 0, 10)
FONT_SECTION_PREVIEW_LEN = 30

ATTRIBUTES_TO_SHOW_IN_SUMMARY_TABLE = [
    'sub_type',
    'base_font',
    'flags',
    'bounding_box',
]


class FontInfo:
    @classmethod
    def extract_font_infos(cls, obj_with_resources: PdfObject) -> ['FontInfo']:
        """
        Extract all the fonts from a given /Resources PdfObject node.
        obj_with_resources must have '/Resources' because that's what _cmap module expects
        """
        resources = obj_with_resources[RESOURCES]

        if isinstance(resources, IndirectObject):
            resources = resources.get_object()

        fonts = resources.get(FONT)

        if fonts is None:
            log.info(f'No fonts found in {obj_with_resources}')
            return []

        fonts = fonts.get_object()
        return [cls.build(label, font, obj_with_resources) for label, font in fonts.items()]

    @classmethod
    def build(cls, label: str, font_ref: IndirectObject, obj_with_resources) -> 'FontInfo':
        """Build a FontInfo object from a IndirectObject ref to a /Font"""
        font_obj = font_ref.get_object()
        font_descriptor = None
        font_file = None

        if font_obj.get(TYPE) != FONT:
            raise TypeError(f"{TYPE} of {font_ref} is not {FONT}")

        if FONT_DESCRIPTOR in font_obj:
            font_descriptor = font_obj[FONT_DESCRIPTOR].get_object()
            font_file_keys = [k for k in font_descriptor.keys() if FONT_FILE in k]

            # There's /FontFile, /FontFile2, etc. but whichever it is there should be only one
            if len(font_file_keys) > 1:
                raise RuntimeError(f"Too many /FontFile keys in {font_descriptor}: {font_file_keys}")
            elif len(font_file_keys) == 0:
                log.info(f"No font_file found in {font_descriptor}")
            else:
                font_file = font_descriptor[font_file_keys[0]].get_object()

        return cls(label, font_ref.idnum, font_obj, font_descriptor, font_file, obj_with_resources)

    def __init__(self, label, idnum, font, font_descriptor, font_file, obj_with_resources):
        self.label = label
        self.idnum = idnum
        self.font_file = font_file
        self.descriptor = font_descriptor

        # /Font attributes
        self.font = font
        self.sub_type = font.get(SUBTYPE)
        self.widths = font.get(WIDTHS) or font.get(W)

        if isinstance(self.widths, IndirectObject):
            self.widths = self.widths.get_object()

        self.base_font = font.get('/BaseFont')
        self.first_and_last_char = [font.get('/FirstChar'), font.get('/LastChar')]
        self.display_title = f"{self.idnum}. Font {self.label} "

        if self.sub_type is None:
            log.warning(f"Font type not given for {self.display_title}")
            self.display_title += "(UNKNOWN FONT TYPE)"
        else:
            self.display_title += f"({self.sub_type[1:]})"

        # FontDescriptor attributes
        if font_descriptor is not None:
            self.bounding_box = font_descriptor.get('/FontBBox')
            self.flags = font_descriptor.get('/Flags')
        else:
            self.bounding_box = None
            self.flags = None

        # /FontFile attributes
        if font_file is not None:
            self.lengths = [font_file[k] for k in FONT_LENGTHS if k in font_file]
            self.stream_data = font_file.get_data()
            self.advertised_length = sum(self.lengths)
            scanner_label = Text(self.display_title, get_label_style(FONT_FILE))
            self.binary_scanner = BinaryScanner(self.stream_data, self, scanner_label)
            self.prepared_char_map = prepare_cm(font) if TO_UNICODE in font else None
            # TODO: shouldn't we be passing ALL the widths?
            self._char_map = build_char_map(label, self.widths[0], obj_with_resources)

            try:
                self.character_mapping = self._char_map[3]
            except (IndexError, TypeError):
                log.warning(f"Exception trying to get character mapping for {self}")
                self.character_mapping = []
        else:
            self.lengths = None
            self.stream_data = None
            self.advertised_length = None
            self.binary_scanner = None
            self.prepared_char_map = None
            self._char_map = None
            self.character_mapping = None

    def width_stats(self):
        if self.widths is None:
            return {}

        return {
            'min': min(self.widths),
            'max': max(self.widths),
            'count': len(self.widths),
            'unique_count': len(set(self.widths)),
        }

    def print_summary(self):
        """Prints a table of info about the font drawn from the various PDF objects. quote_type of None means all."""
        print_section_subheader(str(self), style='font.title')
        #console.print(Panel(self.display_title, width=subheading_width(), padding=(1, 1)), style='font.title')
        console.print(self._summary_table())
        console.line()
        self.print_character_mapping()
        self.print_prepared_charmap()
        console.line()

    def print_character_mapping(self):
        """Prints the character mapping extracted by PyPDF2._charmap in tidy columns"""
        if self.character_mapping is None or len(self.character_mapping) == 0:
            log.info(f"No character map found in {self}")
            return

        print_headline_panel(f"{self} {CHARMAP_TITLE}", style='charmap.title')
        charmap_entries = [_format_charmap_entry(k, v) for k, v in self.character_mapping.items()]

        charmap_columns = Columns(
            charmap_entries,
            column_first=True,
            padding=CHARMAP_PADDING,
            equal=True,
            align='right')

        console.print(Padding(charmap_columns, CHARMAP_TITLE_PADDING), width=subheading_width())
        console.line()

    def print_prepared_charmap(self):
        """Prints the prepared_charmap returned by PyPDF2"""
        if self.prepared_char_map is None:
            log.info(f"No prepared_charmap found in {self}")
            return

        headline = f"{self} Adobe PostScript charmap prepared by PyPDF2"
        print_headline_panel(headline, style='charmap.prepared_title')
        print_bytes(self.prepared_char_map, style='charmap.prepared')
        console.print('')

    def preview_bytes_at_advertised_lengths(self):
        """Show the bytes at the boundaries provided by /Length1, /Length2, and /Length3, if they exist"""
        lengths = self.lengths or []

        if self.lengths is None or len(lengths) <= 1:
            console.print("No length demarcations to preview.", style='grey.dark')

        for i, demarcation in enumerate(lengths[1:]):
            console.print(f"{self.font_file} at /Length{i} ({demarcation}):")
            print(f"\n  Stream before: {self.stream_data[demarcation - FONT_SECTION_PREVIEW_LEN:demarcation + 1]}")
            print(f"\n  Stream after: {self.stream_data[demarcation:demarcation + FONT_SECTION_PREVIEW_LEN]}")

        print(f"\nfinal bytes back from {self.stream_data.lengths[2]} + 10: {self.stream_data[-10 - -f.lengths[2]:]}")

    def _summary_table(self):
        """Build a Rich Table with important info about the font"""
        table = Table('', '', show_header=False)
        table.columns[0].style = 'font.property'
        table.columns[0].justify = 'right'

        def add_table_row(name, value):
            table.add_row(name, Text(str(value), get_type_style(type(value))))

        for attr in ATTRIBUTES_TO_SHOW_IN_SUMMARY_TABLE:
            attr_value = getattr(self, attr)
            add_table_row(attr, attr_value)

        add_table_row('/Length properties', self.lengths)
        add_table_row('total advertised length', self.advertised_length)

        if self.binary_scanner is not None:
            add_table_row('actual length', self.binary_scanner.stream_length)

        if self.prepared_char_map is not None:
            add_table_row('prepared charmap length', len(self.prepared_char_map))

        if self._char_map is not None:
            add_table_row('character mapping count', len(self.character_mapping))

        if self.widths is not None:
            for k, v in self.width_stats().items():
                add_table_row(f"char width {k}", v)

        if self.widths is not None:
            if len(set(self.widths)) == 1:
                table.add_row(
                    'char widths',
                    Text(f"{self.widths[0]} (single value repeated {len(self.widths)} times)", style=get_type_style(list)))
            else:
                add_table_row('char widths', self.widths)
                add_table_row('char widths(sorted)', sorted(self.widths))

        col_0_width = max([len(entry) for entry in table.columns[0]._cells]) + 4
        table.columns[1].max_width = subheading_width() - col_0_width - 3
        return table

    def __str__(self) -> str:
        return self.display_title


def _format_charmap_entry(k: str, v: str) -> Text:
    key = pp.pformat(k)

    for quote_char in ['"', "'"]:
        if len(key) > 1 and key.startswith(quote_char) and key.endswith(quote_char):
            key = key[1:-1]

    return quoted_text(key, 'charmap.byte') + Text(' => ') + quoted_text(str(v), 'charmap.char')


def quoted_text(_string, style: Union[str, None]=None, quote_char_style='white', quote_char="'") -> Text:
    quote_char = Text(quote_char, style=quote_char_style)
    txt = quote_char + Text(_string, style=style or '') + quote_char
    txt.justify = 'center'
    return txt
